@@ -20,14 +20,20 @@ These route factories are used to create all the routes for
 """
 from typing import Optional, List
 from sqlalchemy.orm import sessionmaker
-from sqlalchemy.exc import IntegrityError
+from sqlalchemy.exc import IntegrityError, OperationalError
 from fastapi import status, Depends, Body, HTTPException
 from functools import wraps
 import inspect
 import logging
-from chapps.rest.dbsession import sql_engine
-from chapps.rest.models import AssocOperation, DeleteResp, TextResp
-from chapps.rest.dbmodels import JoinAssoc
+from chapps.dbsession import sql_engine
+from chapps.models import (
+    CHAPPSModel,
+    CHAPPSResponse,
+    AssocOperation,
+    DeleteResp,
+    TextResp,
+)
+from chapps.dbmodels import JoinAssoc
 import chapps.logging
 
 logger = logging.getLogger(__name__)
@@ -38,12 +44,12 @@ async def list_query_params(
     skip: Optional[int] = 0,
     limit: Optional[int] = 1000,
     q: Optional[str] = "%",
-):
+) -> dict:
     """FastAPI dependency for list queries"""
     return dict(q=q, skip=skip, limit=limit)
 
 
-def model_name(cls):
+def model_name(cls) -> str:
     """Convenience function to get the lowercase name of a model"""
     return cls.__name__.lower()
 
@@ -57,11 +63,11 @@ def load_model_with_assoc(cls, assoc: List[JoinAssoc], engine=sql_engine):
     return a closure which can work in any context, it does not return a
     coroutine but a standard synchronous closure.
 
-    :param ~chapps.rest.models.CHAPPSModel cls: a data model class
+    :param ~.CHAPPSModel cls: a data model class
     :param assoc: a list of associations (as
       :class:`~.JoinAssoc` objects)
     :param Optional[~sqlalchemy.engine.Engine] engine: defaults to
-      :const:`chapps.rest.dbsession.sql_engine` if not specified
+      :const:`chapps.dbsession.sql_engine` if not specified
     :rtype: callable
     :returns: a closure as follows:
 
@@ -70,7 +76,7 @@ def load_model_with_assoc(cls, assoc: List[JoinAssoc], engine=sql_engine):
         :param int item_id: if non-zero, the ID of the main record
         :param Optional[str] name: if `item_id` is 0, the `name` of the record
           to match.
-        :rtype: Tuple[~chapps.rest.models.CHAPPSModel, Dict[str, List[~chapps.rest.models.CHAPPSModel]], List[str]]
+        :rtype: Tuple[~.CHAPPSModel, Dict[str, List[~.CHAPPSModel]], List[str]]
         :returns: a :obj:`tuple` containing:
 
           1. the object loaded by ID or name
@@ -84,53 +90,170 @@ def load_model_with_assoc(cls, assoc: List[JoinAssoc], engine=sql_engine):
     mname = model_name(cls)
     assoc_s = "_".join([a.assoc_name for a in assoc])
     fname = f"load_{mname}_with_{assoc_s}"
-    Session = sessionmaker(engine)
+    # Session = sessionmaker(engine)
 
+    @db_wrapper(cls=cls, engine=engine)
     def get_model_and_assoc(item_id: int, name: Optional[str]):
         remarks = []
         items = {k: None for k in [mname, *[a.assoc_name for a in assoc]]}
-        with Session() as sess:
+        # session is a global provided by the decorator
+        if item_id:
+            items[mname] = session.scalar(cls.select_by_id(item_id))
+        if name and not items[mname]:
+            items[mname] = session.scalar(cls.select_by_name(name))
+            if items[mname] and item_id:
+                remarks.append(
+                    f"Selecting {mname} {items[mname].name} with "
+                    f"id {items[mname].id} by name because "
+                    "provided id does not exist."
+                )
+        if not items[mname] and not (item_id or name):
+            logger.debug(  # log this, as it is weird
+                f"{fname}({item_id!r}, {name!r}): unable to load {mname}"
+            )
+            raise HTTPException(  # describe error to the caller
+                status_code=status.HTTP_400_BAD_REQUEST,
+                detail=(
+                    f"One of {cls.id_name()} or name must be provided. "
+                    "If both are provided, {cls.id_name()} is preferred."
+                ),
+            )
+        if items[mname]:
+            for a in assoc:
+                items[a.assoc_name] = getattr(items[mname], a.assoc_name)
+        else:
+            detail = "No {mname} could be found with "
             if item_id:
-                items[mname] = sess.scalar(cls.select_by_id(item_id))
-            if name and not items[mname]:
-                items[mname] = sess.scalar(cls.select_by_name(name))
-                if items[mname] and item_id:
-                    remarks.append(
-                        f"Selecting {mname} {items[mname].name} with "
-                        f"id {items[mname].id} by name because "
-                        "provided id does not exist."
-                    )
-            if not items[mname] and not (item_id or name):
-                logger.debug(  # log this, as it is weird
-                    f"{fname}({item_id!r}, {name!r}): unable to load {mname}"
-                )
-                raise HTTPException(  # describe error to the caller
-                    status_code=status.HTTP_400_BAD_REQUEST,
-                    detail=(
-                        f"One of {cls.id_name()} or name must be provided. "
-                        "If both are provided, {cls.id_name()} is preferred."
-                    ),
-                )
-            if items[mname]:
-                for a in assoc:
-                    items[a.assoc_name] = getattr(items[mname], a.assoc_name)
-            else:
-                detail = "No {mname} could be found with "
-                if item_id:
-                    detail += "id {item_id}"
-                    if name:
-                        detail += f" or with "
+                detail += "id {item_id}"
                 if name:
-                    detail += "name {name}"
-                raise HTTPException(
-                    status_code=status.HTTP_404_NOT_FOUND, detail=detail
-                )
+                    detail += f" or with "
+            if name:
+                detail += "name {name}"
+            raise HTTPException(
+                status_code=status.HTTP_404_NOT_FOUND, detail=detail
+            )
         return (items.pop(mname), items, remarks)
 
     get_model_and_assoc.__name__ = fname
     return get_model_and_assoc
 
 
+def load_models_with_assoc(
+    cls: CHAPPSModel, *, assoc: JoinAssoc, engine=sql_engine
+) -> callable:
+    """Build a map of source name => associated object id
+
+    :param cls: source model
+
+    :param assoc: a join association representing the associated model
+
+    :param engine: override the SQLA engine if desired
+
+    :returns: a mapper function which accepts a list of IDs of the source model
+      and returns a list of dicts with `<source_model>_name` and
+      `<assoc_model>_id` fields, mapping the source objects onto the IDs of
+      their associated objects of the configured type.
+
+    .. todo::
+
+      Employ eager loading on the target association.
+
+    """
+    mname = model_name(cls)
+    fname = f"eager_load_{mname}_with_{assoc.assoc_name}"
+    # Session = sessionmaker(engine)
+
+    @db_wrapper(cls=cls, engine=engine)
+    def map_model_names_to_assoc(
+        item_ids: List[int],  # name_tail: Optional[str] = None
+    ):
+        eager_loaded_models = session.scalars(
+            cls.select_by_ids(
+                item_ids, getattr(cls.Meta.orm_model, assoc.assoc_name)
+            )
+        )
+        return list(eager_loaded_models)
+
+    map_model_names_to_assoc.__name__ = fname
+    return map_model_names_to_assoc
+
+
+def db_wrapper(  # a decorator with parameters
+    *,
+    cls,
+    engine=sql_engine,
+    exception_message: str = ("{route_name}:{model}"),
+    empty_set_message: str = ("Unable to find a matching {model}"),
+):
+    """Decorator for database interactions
+
+    :param ~chapps.models.CHAPPSModel cls: the data model class
+
+    :param ~sqlalchemy.engine.Engine engine: an :mod:`SQLAlchemy` engine, which
+      defaults to the package-wide one declared in
+      :mod:`~chapps.dbsession`
+
+    :param str exception_message: a message to include if any untrapped
+      exception occurs; defaults to ``{route_name}:{model}``.  Only those two
+      symbols are available for expansion.  All arguments are appended.
+
+    :param str empty_set_message: included if a SELECT results in an empty set;
+      defaults to ``Unable to find a matching {model}`` and supports both
+      substitutions that `exception_message` does
+
+    :returns: a `decorator`_ closure, which will be called with the function to
+      be `decorated`_ as its argument.  This is a regular callable decorator.
+
+    :rtype: callable which wraps and returns a function
+
+    The decorator sets up some global symbols for use inside the DB access function:
+
+      :session: a :class:`~sqlalchemy.orm.Session` instance created in a
+        context containing the execution of the wrapped coroutine, suitable for
+        performing database interactions, and which will be automatically
+        closed after the coroutine completes
+
+      :model_name: a string containing the lowercase name of the model
+
+    .. _decorator: https://docs.python.org/3/glossary.html#term-decorator
+    .. _decorated: https://peps.python.org/pep-0318/
+    .. _coroutine: https://docs.python.org/3/library/asyncio-task.html#coroutines
+
+    """
+
+    mname = model_name(cls)
+    Session = sessionmaker(engine)
+
+    def db_func_wrapper(db_func):
+        exc = exception_message.format(
+            route_name=db_func.__name__, model=mname
+        )
+        empty = empty_set_message.format(
+            route_name=db_func.__name__, model=mname
+        )
+
+        @wraps(db_func)
+        def wrapped_interaction(*args, **kwargs):
+            with Session() as session:
+                db_func.__globals__["session"] = session
+                db_func.__globals__["model_name"] = mname
+
+                try:
+                    result = db_func(*args, **kwargs)
+                    if result:
+                        return result
+                except HTTPException as e:
+                    raise e
+                except Exception:
+                    logger.exception(exc + f"({args!r},{kwargs!r})")
+            raise HTTPException(status_code=404, detail=empty)
+
+        wrapped_interaction.__doc__ = db_func.__doc__
+        return wrapped_interaction  # a regular function
+
+    return db_func_wrapper  # a regular function
+
+
 def db_interaction(  # a decorator with parameters
     *,
     cls,
@@ -140,11 +263,11 @@ def db_interaction(  # a decorator with parameters
 ):
     """Decorator for database interactions
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the data model class
+    :param ~chapps.models.CHAPPSModel cls: the data model class
 
     :param ~sqlalchemy.engine.Engine engine: an :mod:`SQLAlchemy` engine, which
       defaults to the package-wide one declared in
-      :mod:`~chapps.rest.dbsession`
+      :mod:`~chapps.dbsession`
 
     :param str exception_message: a message to include if any untrapped
       exception occurs; defaults to ``{route_name}:{model}``.  Only those two
@@ -200,6 +323,8 @@ def db_interaction(  # a decorator with parameters
                     result = await rt_coro(*args, **kwargs)
                     if result:
                         return result
+                except HTTPException as e:
+                    raise e
                 except Exception:
                     logger.exception(exc + f"({args!r},{kwargs!r})")
             raise HTTPException(status_code=404, detail=empty)
@@ -219,13 +344,13 @@ def get_item_by_id(
 ):
     """Build a route coroutine to get an item by ID
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the main data model for the request
+    :param ~chapps.models.CHAPPSModel cls: the main data model for the request
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
+    :param ~chapps.models.CHAPPSResponse response_model: the response
       model
 
     :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+      :const:`~chapps.dbsession.sql_engine`
 
     :param List[~chapps.rest.dbmodels.JoinAssoc] assoc: if included, these
       associations will be included as optional keys in the response
@@ -287,14 +412,14 @@ def get_item_by_id(
 def list_items(cls, *, response_model, engine=sql_engine):
     """Build a route coroutine to list items
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the main data model for the
+    :param ~chapps.models.CHAPPSModel cls: the main data model for the
       request
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
+    :param ~chapps.models.CHAPPSResponse response_model: the response
       model
 
     :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+      :const:`~chapps.dbsession.sql_engine`
 
     The returned closure expects to receive the query parameters as a dict,
     since that is what the dependency will return.  Its signature is
@@ -329,19 +454,22 @@ def list_items(cls, *, response_model, engine=sql_engine):
 
 
 def list_associated(
-    cls, *, assoc: JoinAssoc, response_model, engine=sql_engine
+    cls: CHAPPSModel,
+    *,
+    assoc: JoinAssoc,
+    response_model: CHAPPSResponse,
+    engine=sql_engine,
 ):
     """Build a route to list associated items with pagination
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the main data model
+    :param cls: the main data model
 
-    :param ~chapps.rest.dbmodels.JoinAssoc assoc: the association to list
+    :param assoc: the association to list
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
-      model
+    :param response_model: the response model
 
-    :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+    :param sqlalchemy.engine.Engine engine: defaults to
+      :const:`~chapps.dbsession.sql_engine`
 
     The returned coroutine will paginate a list of the associated objects,
     given the ID of a main (source) object to use to select associations.  The
@@ -357,41 +485,21 @@ def list_associated(
     the associated object, goverened by the search and window parameters in
     `qparams`.
 
-    .. todo::
-
-      Simplify/clarify signature of inner coroutine -- it will always be the
-      same: `(item_id: int, qparams: dict = Depends(list_query_params))`
-
     """
     mname = model_name(cls)
     fname = f"{mname}_list_{assoc.assoc_name}"
-    params = dict(item_id=int, qparams=dict)
-    p_dfls = dict(item_id=None, qparams=Depends(list_query_params))
 
     @db_interaction(cls=cls, engine=engine)
-    async def assoc_list(*pargs, **args):
-        # TODO: signature can be simplified
-        # item_id for source object of type cls
-        # the specified association is listed according to qparams
+    async def assoc_list(
+        item_id: int, qparams: dict = Depends(list_query_params)
+    ):
         stmt = assoc.assoc_model.windowed_list_by_ids(
-            subquery=assoc.select_ids_by_source_id(args["item_id"]),
-            **args["qparams"],
+            subquery=assoc.select_ids_by_source_id(item_id), **qparams
         )
         assocs = assoc.assoc_model.wrap(session.scalars(stmt))
         if assocs:
             return response_model.send(assocs)
 
-    routeparams = [
-        inspect.Parameter(
-            name=param,
-            kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
-            annotation=type_,
-            default=p_dfls[param],
-        )
-        for param, type_ in params.items()
-    ]
-    assoc_list.__signature__ = inspect.Signature(routeparams)
-    assoc_list.__annotations__ = params
     assoc_list.__name__ = fname
     assoc_list.__doc__ = (
         f"List **{assoc.assoc_name}** associated with a particular **{mname}**,"
@@ -404,11 +512,11 @@ def list_associated(
 def delete_item(cls, *, response_model=DeleteResp, engine=sql_engine):
     """Build a route coroutine to delete an item by ID
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the data model to manage
+    :param ~chapps.models.CHAPPSModel cls: the data model to manage
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: defaults to :class:`~chapps.rest.models.DeleteResp`
+    :param ~chapps.models.CHAPPSResponse response_model: defaults to :class:`~chapps.models.DeleteResp`
 
-    :param ~sqlalchemy.engine.Engine engine: defaults to :const:`~chapps.rest.dbsession.sql_engine`
+    :param ~sqlalchemy.engine.Engine engine: defaults to :const:`~chapps.dbsession.sql_engine`
 
     The returned coroutine accepts a list of record IDs for the specified
     object type and delete them.  Its signature is:
@@ -451,19 +559,19 @@ def adjust_associations(
 ):
     """Build a route to add or subtract association lists, or set unitary ones
 
-    :param ~chapps.rest.models.CHAPPSModel cls: a data model class
+    :param ~chapps.models.CHAPPSModel cls: a data model class
 
     :param List[~chapps.rest.dbmodels.JoinAssoc] assoc: list of associations to
       operate on
 
-    :param ~chapps.rest.models.AssocOperation assoc_op: operation to perform on
+    :param ~chapps.models.AssocOperation assoc_op: operation to perform on
       the association
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
+    :param ~chapps.models.CHAPPSResponse response_model: the response
       model to send
 
     :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+      :const:`~chapps.dbsession.sql_engine`
 
     The returned coroutine provides logic for a route which adds or subtracts
     elements to or from those already associated with the main object.  Its
@@ -561,15 +669,15 @@ def update_item(
 ):
     """Build a route to update items.
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the main data model
+    :param ~chapps.models.CHAPPSModel cls: the main data model
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
+    :param ~chapps.models.CHAPPSResponse response_model: the response
       model
 
     :param ~chapps.rest.dbmodels.JoinAssoc assoc: the association to list
 
     :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+      :const:`~chapps.dbsession.sql_engine`
 
     The returned coroutine implements an API route for updating an item by ID,
     optionally including any associations included when the route coroutine is
@@ -681,7 +789,7 @@ def update_item(
     update_i.__name__ = fname
     update_i.__doc__ = f"""
         Update a **{mname}** record by ID.<br/>
-        All **{mname}** attributes are required.<br/>
+        Only the **id** and the adjusted attribute(s) are required.<br/>
         Associations are not required, but if provided (by ID), will
         completely replace any existing association relationships
         of the same type.
@@ -690,13 +798,19 @@ def update_item(
 
 
 def create_item(
-    cls, *, response_model, params=None, assoc=None, engine=sql_engine
+    cls,
+    *,
+    response_model,
+    params=None,
+    defaults=None,
+    assoc=None,
+    engine=sql_engine,
 ):
     """Build a route coroutine to create new item records
 
-    :param ~chapps.rest.models.CHAPPSModel cls: the main data model
+    :param ~chapps.models.CHAPPSModel cls: the main data model
 
-    :param ~chapps.rest.models.CHAPPSResponse response_model: the response
+    :param ~chapps.models.CHAPPSResponse response_model: the response
       model
 
     :param dict params: defaults to ``dict(name=str)``; specify to provide
@@ -704,11 +818,11 @@ def create_item(
       models currently are expected to have a `name` column, which is not
       allowed to be null.
 
-    :param ~chapps.rest.dbmodels.JoinAssoc assoc: the associations to attach,
+    :param ~.JoinAssoc assoc: the associations to attach,
       if any
 
     :param ~sqlalchemy.engine.Engine engine: defaults to
-      :const:`~chapps.rest.dbsession.sql_engine`
+      :const:`~chapps.dbsession.sql_engine`
 
     The returned coroutine implements an API route for creating an item,
     setting all its elements (other than ID) to whatever values are provided.
@@ -719,12 +833,12 @@ def create_item(
 
     """
     params = params or dict(name=str)
+    defaults = defaults or dict()
     mname = model_name(cls)
     fname = f"create_{mname}"
 
     @db_interaction(cls=cls, engine=engine)
     async def create_i(*pargs, **args):
-        """Bogus create item docstring"""
         extras = {}
         assoc_ret = {}
         if assoc:
@@ -742,6 +856,10 @@ def create_item(
                 status_code=status.HTTP_409_CONFLICT,
                 detail=f"Unique key conflict creating {mname}.",
             )
+        except OperationalError as e:
+            raise HTTPException(
+                status_code=status.HTTP_400_BAD_REQUEST, detail=str(e)
+            )
         except Exception:
             logger.exception(f"{fname}: {args!r}")
             raise HTTPException(
@@ -773,7 +891,7 @@ def create_item(
         inspect.Parameter(
             name=param,
             kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
-            default=Body(...),
+            default=Body(defaults.get(param, ...)),
             annotation=type_,
         )
         for param, type_ in params.items()
@@ -790,14 +908,25 @@ def create_item(
                 for a in assoc
             ]
         )
+    defaulted_attrs = [a for a, d in defaults.items() if d != ...]
+    if len(defaulted_attrs):
+        dfl_attr_doc = (
+            "The `name` attribute is required, along with any others"
+            + " not described below as optional.<br/>"
+            + "The following attributes are <b>optional</b>:<br/>"
+            + f"<ul>{' '.join(['<li>`'+e+'`</li>' for e in defaulted_attrs])}"
+            + "</ul>"
+        )
+    else:
+        dfl_attr_doc = "All attributes are required."
     create_i.__signature__ = inspect.Signature(routeparams)
     create_i.__annotations__ = params
     create_i.__name__ = fname
-    create_i.__doc__ = f"""
-        Create a new **{mname}** record in the database.<br/>
-        All attributes are required.<br/>
-        The new object will be returned, including its ID.<br/>
-        Raises descriptive errors on 409; checking the detail
-          of the error may aid in debugging.
-        """
+    create_i.__doc__ = (
+        f"Create a new **{mname}** record in the database.<br/>"
+        + "The new object will be returned, including its ID.  "
+        + "Raises descriptive errors on 409; checking the detail "
+        + "of the error may aid in debugging.<br/>"
+        + dfl_attr_doc
+    )
     return create_i
